@@ -3,9 +3,8 @@ Feature-Based Methods Module
 """
 
 from ._fe import FE
-from ._coral import CORAL, DeepCORAL
-from ._adda import ADDA
-from ._dann import DANN
+from ._coral import CORAL
 from ._msda import mSDA
+from ._deep import DANN, ADDA, DeepCORAL
 
 __all__ = ["FE", "CORAL", "DeepCORAL", "ADDA", "DANN", "mSDA"]
